/* $Header: /home/cvsroot/MyPasswordSafe/src/safe.cpp,v 1.29 2004/01/26 09:36:05 nolan Exp $
 * Copyright (c) 2004, Semantic Gap Solutions
 * All rights reserved.
 *   
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided
 * that the following conditions are met:
 *  -  Redistributions of source code must retain the
 *     above copyright notice, this list of conditions
 *     and the following disclaimer.
 *  -  Redistributions in binary form must reproduce the
 *     above copyright notice, this list of conditions and
 *     the following disclaimer in the documentation and/or
 *     other materials provided with the distribution.
 *  -  Neither the name of Semantic Gap Solutions nor the
 *     names of its contributors may be used to endorse or
 *     promote products derived from this software without
 *     specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#include <qlistview.h>
#include <string>
#include <map>
#include <algorithm>
#include <iostream>
#include <fstream>
#include "securedstring.hpp"
#include "safe.hpp"
#include "serializers.hpp"
#include "myutil.hpp"
#include "pwsafe/Util.h"

using namespace std;

SafeSerializer::SerializerMap SafeSerializer::m_serializers;
PlainTextLizer _plain_text_lizer;
BlowfishLizer _blowfish_lizer;

SafeItem::SafeItem()
  : m_name(""), m_user(""), m_notes("")
{
}

SafeItem::SafeItem(const string &name, const string &user,
		   const EncryptedString &password, const string &notes)
  : m_name(name), m_user(user), m_notes(notes), m_password(password)
{
}

void SafeItem::setName(const string &name)
{
  m_name = name;
}

void SafeItem::setUser(const string &user)
{
  m_user = user;
}

void SafeItem::setPassword(const EncryptedString &password)
{
  //m_password.setAlgorithm(password.getAlgorithm());
  m_password.set(password);
}

void SafeItem::setPassword(const char *password)
{
  m_password.set(password);
}

void SafeItem::setNotes(const string &notes)
{
  m_notes = notes;
}

void SafeItem::updateListItem(QListViewItem *item)
{
  // i wish setText took string
  item->setText(0, QString::fromUtf8(getName()));
  item->setText(1, QString::fromUtf8(getUser()));
  item->setText(2, QString::fromUtf8(getNotes()));

  // FIXME: when the safe gets refactored to use a view,
  // remeber to set the item's height
  string notes(getNotes());
  int lines = split(notes, '\n').size();
  if(lines > 1)
    item->setMultiLinesEnabled(true);
}

Safe::Safe(QListView *listbox)
  : m_path(""), m_type(""), m_passphrase(NULL), m_listbox(listbox),
    m_changed(false)
{
}

struct SafeCleaner
{
  void operator () (Safe::ItemMap::reference r)
  {
    if(r.first != NULL) {
      delete r.first;
    }
  }
};

Safe::~Safe()
{
  empty();
}

bool Safe::checkPassword(const string &path, const SecuredString &password)
{
  string type(getExtension(path));
  if(type.empty())
    return false;
  return checkPassword(path, type, password);
}

bool Safe::checkPassword(const string &path, const string &ext, const SecuredString &password)
{
  SafeSerializer *serializer = SafeSerializer::create(ext);
  if(serializer) {
    return serializer->checkPassword(path, password);
  }
  return false;
}

Safe *Safe::load(const string &path, const string &type, const EncryptedString &passphrase, const string &def_user, QListView *listbox)
{
  SafeSerializer *serializer = SafeSerializer::create(type);
  if(serializer) {
    Safe *safe = new Safe(listbox);
    if(safe != NULL) {
      if(serializer->load(*safe, path, passphrase, def_user)) {
	safe->setPath(path);
	safe->setType(type);
	safe->setPassPhrase(passphrase);
	safe->setChanged(false);
	return safe;
      }
      else {
	delete safe;
      }
    }
  }
  return NULL;
}

Safe *Safe::load(const string &path, const EncryptedString &passphrase, const string &def_user, QListView *listbox)
{
  string type(getExtension(path));
  if(type.empty())
    return NULL;
  else
    return load(path, type, passphrase, def_user, listbox);
}

bool Safe::save(const string &path, const string &type,
		const EncryptedString &passphrase, const string &def_user)
{
  SafeSerializer *serializer = SafeSerializer::create(type);
  if(serializer) {
    setPath(path);
    setType(type);

    // setPassPhrase can't set it back to itself because
    // the references all go back to m_passphrase. So
    // we do a quick check to avert the problems that
    // that was causing.
    if(passphrase != getPassPhrase())
      setPassPhrase(passphrase);

    // FIXME: make this an option
    makeBackup(path.c_str());

    if(serializer->save(*this, path, def_user)) {
      setChanged(false);
      return true;
    }
  }
  return false;
}

bool Safe::save(const string &path, const EncryptedString &passphrase, const string &def_user)
{
  string type(getExtension(path));
  if(type.empty())
    return false;
  else
    return save(path, type, passphrase, def_user);
}

bool Safe::save(const string &def_user)
{
  if(m_path.length() > 0 && m_passphrase.length() > 0) {
    if(save(m_path, m_passphrase, def_user))
      return true;
  }
  return false;
}

void Safe::setPath(const string &path)
{
  m_path = path;
}

void Safe::setType(const string &type)
{
  // NOTE: we don't check type because only Safe::save calls it, and
  // it used type to create a serializer so it's good
  m_type = type;
}

void Safe::setPassPhrase(const EncryptedString &phrase)
{
  m_passphrase.set(phrase);
}

void Safe::setChanged(bool value)
{
  m_changed = value;
}

SafeItem *Safe::findItem(QListViewItem *item)
{
  if(m_listbox != NULL) {
    ItemMap::iterator i = m_items.find(item);
    if(i != m_items.end())
      return &(i->second);
  }
  return NULL;
}

void Safe::addItem(SafeItem &item)
{
  if(m_listbox != NULL) {
    QListViewItem *i = new QListViewItem(m_listbox);//, item.getName(), item.getUser(), item.getNotes());
    if(i) {
      m_items.insert(make_pair(i, item));
      item.updateListItem(i);
      setChanged(true);
    }
  }
}

bool Safe::delItem(QListViewItem *item)
{
  ItemMap::iterator i = m_items.find(item);
  if(i != m_items.end()) {
    if(i->first != NULL)
      delete i->first; // this is the QListViewItem pointer
    m_items.erase(i);
    setChanged(true);
    return true;
  }
  return false;
}

void Safe::empty()
{
/*
  if(!m_items.empty())
    for_each(m_items.begin(), m_items.end(), SafeCleaner());
*/
  for(ItemMap::iterator i = m_items.begin();
    i != m_items.end();
    i++) {
    if(i->first != NULL) {
      // FIXME: i->first is CONST!!
      delete i->first;
    }
  }
  m_items.erase(m_items.begin(), m_items.end());
}

bool Safe::makeBackup(const char *path)
{
  if(path == NULL)
    return false;

  FILE *in = fopen(path, "rb");
  if(in != NULL) {
    string new_path(path);
    new_path += "~";
    FILE *out = fopen(new_path.c_str(), "wb");
    char buffer[1024];
    int num_read = 0;
    do {
    	num_read = fread(buffer, 1, 1024, in);
	if(num_read != 0)
	  fwrite(buffer, 1, num_read, out);
    } while(num_read > 0);
    fclose(out);
    fclose(in);
    return true;
  }
  return false;
}

SafeSerializer *SafeSerializer::create(const string &type)
{
  if(!m_serializers.empty()) {
    SerializerMap::iterator i = m_serializers.find(type);
    if(i != m_serializers.end())
      return i->second;
  }
  return NULL;
}

bool SafeSerializer::add(SafeSerializer *serializer, const string &ext)
{
  // check to see ext is handled already
  SafeSerializer *s = create(ext);
  if(s != NULL)
    return false;
  m_serializers.insert(make_pair(ext, serializer));
  return true;
}

struct TypeConCater
{
private:
  string result;

public:
  void setResult(const string &s) { result = s; }
  const string &getResult() { return result; }

  void operator () (SafeSerializer::SerializerMap::reference r)
  {
    setResult(getResult() + r.second->description() + "\n");
  }
};

string SafeSerializer::getTypes()
{
  TypeConCater cater;
  return(for_each(m_serializers.begin(),
		  m_serializers.end(),
		  cater).getResult() + "All files (*)\n");
}

const char *SafeSerializer::getExtForType(const char *type)
{
  for(SerializerMap::iterator i = m_serializers.begin();
      i != m_serializers.end();
      i++) {
    if(string(i->second->description()) == string(type)) {
      return i->first.c_str();
    }
  }
  return NULL;
}
