/* $Header: /home/cvsroot/MyPasswordSafe/src/safe.hpp,v 1.26 2004/01/23 08:25:22 nolan Exp $
 * Copyright (c) 2004, Semantic Gap Solutions
 * All rights reserved.
 *   
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided
 * that the following conditions are met:
 *  -  Redistributions of source code must retain the
 *     above copyright notice, this list of conditions
 *     and the following disclaimer.
 *  -  Redistributions in binary form must reproduce the
 *     above copyright notice, this list of conditions and
 *     the following disclaimer in the documentation and/or
 *     other materials provided with the distribution.
 *  -  Neither the name of Semantic Gap Solutions nor the
 *     names of its contributors may be used to endorse or
 *     promote products derived from this software without
 *     specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef SAFE_HPP
#define SAFE_HPP

//#include <qlistview.h>
#include <string>
#include <map>
#include "securedstring.hpp"
#include "encryptedstring.hpp"

class QListView;
class QListViewItem;

using std::string;
using std::map;

class SafeItem
{
public:
  SafeItem();
  SafeItem(const string &name, const string &user,
	   const EncryptedString &password, const string &notes);
    
  inline const char *getName() const { return m_name.c_str(); }
  inline const char *getUser() const { return m_user.c_str(); }
  inline const EncryptedString &getPassword() const { return m_password; }
  inline const char *getNotes() const { return m_notes.c_str(); }
    
  void setName(const string &name);
  void setUser(const string &user);
  void setPassword(const EncryptedString &password);
  void setPassword(const char *password);
  void setNotes(const string &notes);

  void updateListItem(QListViewItem *item);
    
private:
  string m_name, m_user, m_notes;
  EncryptedString m_password;
};

class Safe
{
public:
  typedef map<QListViewItem *, SafeItem> ItemMap;
    
  Safe(QListView *listbox = NULL);
  ~Safe();

  static bool checkPassword(const string &path, const SecuredString &password);
  // checks to see if PASSWORD will open PATH returning true if it does
  static bool checkPassword(const string &path, const string &ext, const SecuredString &password);
  // does the same thing as the above, but allows the file type to be
  // specified by setting EXT to the extension of the file type

  static Safe *load(const string &path, const string &type, const EncryptedString &passphrase, const string &def_user, QListView *listbox = NULL);
  static Safe *load(const string &path, const EncryptedString &passphrase, const string &def_user, QListView *listbox = NULL);

  bool save(const string &path, const string &type,
	    const EncryptedString &passphrase, const string &def_user);
  bool save(const string &path, const EncryptedString &passphrase, const string &def_user);
  bool save(const string &def_user);
    
  inline const string &getPath() const { return m_path; }
  void setPath(const string &path);
  inline const string &getType() const { return m_type; }
  inline const EncryptedString getPassPhrase() const { return m_passphrase.get(); }
  void setPassPhrase(const EncryptedString &phrase);

  inline bool changed() { return m_changed; }
  void setChanged(bool value);

  SafeItem *findItem(QListViewItem *item);
  void addItem(SafeItem &item);
  bool delItem(QListViewItem *item);
  ItemMap &getItems() { return m_items; }
  void empty();

  inline QListView *getListBox() { return m_listbox; }
  inline void setListBox(QListView *l) { m_listbox = l; }

protected:
  bool makeBackup(const char *path);
  void setType(const string &type);

private:
  string m_path, m_type;
  EncryptedString m_passphrase;
  ItemMap m_items;
  QListView *m_listbox;
  bool m_changed;
};

class SafeSerializer
{
public:
  virtual const char *description() = 0;
  // Returns the string to be used in file open/save dialogs

  virtual bool checkPassword(const string &path, const SecuredString &password) = 0;
  // returns true if password can open path, false if it can not

  virtual bool load(Safe &safe, const string &path, const EncryptedString &passphrase, const string &def_user) = 0;
  // loads the file specified by path into safe return true on success,
  // false on failure
  virtual bool save(Safe &safe, const string &path, const string &def_user) = 0;
  // saves safe to the file specified by path. returns true if success,
  // false on failure
    
  static SafeSerializer *create(const string &type);
  // public interface to the SafeSerializer factory. Returns a pointer
  // to a SafeSerializer that handles files with the specified
  // file extension
    
  typedef map<string, SafeSerializer *> SerializerMap;
  static const SerializerMap &serializers() { return m_serializers; }
  // returns the std::map of serializers. The map is a pair made of
  // extension and the pointer to the serializer.
  static string getTypes();
  static const char *getExtForType(const char *type);

protected:
  static bool add(SafeSerializer *serializer, const string &ext);
  // register a SafeSerializer to handle the ext extension with the factory
    
private:
  static SerializerMap m_serializers;
};

#endif
