/* $Header: /home/cvsroot/MyPasswordSafe/src/serializers.cpp,v 1.32 2004/01/26 09:36:05 nolan Exp $
 * Copyright (c) 2004, Semantic Gap Solutions
 * All rights reserved.
 *   
 * Redistribution and use in source and binary forms,
 * with or without modification, are permitted provided
 * that the following conditions are met:
 *  -  Redistributions of source code must retain the
 *     above copyright notice, this list of conditions
 *     and the following disclaimer.
 *  -  Redistributions in binary form must reproduce the
 *     above copyright notice, this list of conditions and
 *     the following disclaimer in the documentation and/or
 *     other materials provided with the distribution.
 *  -  Neither the name of Semantic Gap Solutions nor the
 *     names of its contributors may be used to endorse or
 *     promote products derived from this software without
 *     specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND
 * CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
 * PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
 * IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * BlowfishLizer uses code from Password Safe, and is covered
 * by the Artistic License found in pwsafe/LICENSE
 *
 * FIXME: this file needs i8ning
 */
#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <algorithm>
#include "safe.hpp"
#include "securedstring.hpp"
#include "encryptedstring.hpp"
#include "serializers.hpp"
#include "exception.hpp"
#include "myutil.hpp"
#include "pwsafe/BlowFish.h"
#include "pwsafe/Util.h"

using namespace std;
using boost::shared_ptr;

PlainTextLizer::PlainTextLizer()
{
  if(!add(this, "txt")) //"Plain Text (*.txt)"))
    throw Exception("Error registering PlainTextLizer with the factory");
}

const char *PlainTextLizer::description()
{
  static const char str[] = "Plain Text (*.txt)";
  return str;
}

bool PlainTextLizer::checkPassword(const string &path, const SecuredString &password)
{
  ifstream in(path.c_str());
  string line;
  getline(in, line, '\n');
  in.close();
  if(line == password.get())
    return true;
  else
    return false;
}

bool PlainTextLizer::load(Safe &safe, const string &path, const EncryptedString &passphrase, const string &def_user)
{
  ifstream file(path.c_str());
  if(file.is_open()) {
    string line;
    getline(file, line, '\n');

    if(EncryptedString(line) != passphrase)
      return false; // passphrase was invalid

    while(!file.eof()) {
      getline(file, line, '\n');
      vector<string> items = split(line, '\t');
      if(items.size() != 3) {
	break;
      }
      getline(file, line, '\n');
      SafeItem safe_item(items[0], items[1],
			 SecuredString(items[2].c_str()), line);
      safe.addItem(safe_item);
    }

    file.close();
    return true;
  }
  return false;
}

class PTLSaver
{
  ofstream &m_str;

public:
  PTLSaver(ofstream &str)
    : m_str(str)
  {
  }

  void operator() (Safe::ItemMap::reference r)
  {
    SafeItem &i(r.second);
    // NOTE: password is decrypted
    m_str << i.getName() << '\t' << i.getUser() << '\t'
	  << i.getPassword().get().get() << endl
	  << i.getNotes() << endl;
  }
};

bool PlainTextLizer::save(Safe &safe, const string &path, const string &def_user)
{
  Safe::ItemMap &items(safe.getItems());
  ofstream file(path.c_str());
  if(!file.is_open()) {
    return false;
  }

  // NOTE: the passphrase is decrypted...AND SAVED TO DISK!!
  file << safe.getPassPhrase().get().get() << endl;
  for_each(items.begin(), items.end(), PTLSaver(file));
  file.close();
  return true;
}


BlowfishLizer::BlowfishLizer()
{
  if(!add(this, "dat")) //"Blowfish Safes (*.dat)"))
    cout << "Error registering BlowfishLizer" << endl;
}

const char *BlowfishLizer::description()
{
  static char desc[] = "Password Safes (*.dat)";
  return desc;
}

int BlowfishLizer::writeCBC(FILE *fp, string &data, const SecuredString &passkey,
			    const unsigned char *salt, unsigned char *ipthing)
{
  return _writecbc(fp, (const unsigned char *)data.c_str(), data.length(),
		   (const unsigned char *)passkey.get(), passkey.length(),
		   salt, SaltLength, ipthing);
}

// Adapted from Counterpane's PasswordSafe
int BlowfishLizer::readCBC(FILE *fp, string &data, const SecuredString &passkey, const unsigned char *salt, unsigned char *ipthing)
{
  // We do a double cast because the LPCSTR cast operator is overridden by the CString class
  // to access the pointer we need,
  // but we in fact need it as an unsigned char. Grrrr.
  //LPCSTR passstr = LPCSTR(app.m_passkey);
  const char *passstr = passkey.get();

  unsigned char *buffer = NULL;
  unsigned int buffer_len = 0;
  int retval;

  retval = _readcbc(fp, buffer, buffer_len,
		    (const unsigned char *)passstr, passkey.length(),
		    salt, SaltLength, ipthing);
  if (buffer_len > 0) {
    //CMyString str(LPCSTR(buffer), buffer_len);
    //string str(buffer, buffer_len);
    buffer[buffer_len] = '\0';
    data = (char *)buffer;
    trashMemory(buffer, buffer_len);
    delete[] buffer;
  } else {
    data = "";
  }
  return retval;
}

void trimRight(string &str)
{
  string::iterator i = str.end();
  for(; i != str.begin(); i--) {
    if(!isspace(*i) && *i != '\0')
      break;
  }
  str.erase(i+1, str.end());
}

void trimLeft(string &str)
{
  string::iterator i = str.begin();
  for(; i != str.end(); i++) {
    if(!isspace(*i))
      break;
  }
  str.erase(str.begin(), i);
}

bool BlowfishLizer::checkPassword(const string &path, const SecuredString &password)
{
  FILE *in = fopen(path.c_str(), "rb");
  if(in != NULL) {
    //cout << "opened " << path << endl;
    //cout << "using " << password.get() << " as the password" << endl;
    unsigned char randstuff[StuffSize];
    unsigned char randhash[20];

    memset(randstuff, 0, StuffSize);
    fread(randstuff, 1,  8, in);
    fread(randhash,  1, 20,  in);

    fclose(in);

    unsigned char genhash[20];
    GenRandhash(password, randstuff, genhash);
    /*
      cout << "Randstuff" << endl;
      for(int i = 0; i < 8; i++) {
      printf(" %x", randstuff[i]);
      }
      cout << endl << "Genhash" << endl;
      for(int i = 0; i < 20; i++) {
      printf(" %x", genhash[i]);
      }
      cout << endl << "Randhash" << endl;
      for(int i = 0; i < 20; i++) {
      printf(" %x", randhash[i]);
      }
      cout << endl;
    */
    if(memcmp(randhash, genhash, 20) == 0) {
      return true;
    }
  }
  //cout << "failed to open " << path << endl;
  return false;
}

int BlowfishLizer::readHeader(FILE *in, unsigned char randstuff[8], unsigned char randhash[20],
			      unsigned char salt[SaltLength], unsigned char ipthing[8])
{
  int numRead = 0;
  numRead += fread(randstuff, 1,  8, in);
  numRead += fread(randhash,  1, 20,  in);
  numRead += fread(salt,    1, SaltLength, in);
  numRead += fread(ipthing, 1, 8, in); 
  return numRead;
}

int BlowfishLizer::readEntry(FILE *in, SafeItem &item, EncryptedString &es, unsigned char *ipthing, const string &def_user)
{
  SecuredString data;
  string tempdata;

  int numread = 0;
  numread += es.read(in, ipthing);
  if(es.length() > 0) {
    data.set(es.get());
    tempdata = data.get();

    vector<string> name_user(split(tempdata, '\xAD'));
    if(name_user.size() == 2) {
      trimRight(name_user[0]);
      item.setName(name_user[0]);
      trimLeft(name_user[1]);
      item.setUser(name_user[1]);
    }
    // check for the other special character
    else {
      name_user = split(tempdata, '\xA0');
      if(name_user.size() == 2) {
	trimLeft(name_user[0]);
	item.setName(name_user[0]);
	// FIXME: set user name to the default user name
	item.setUser(def_user);
      }
      else {
	item.setName(tempdata);
	item.setUser("");
      }
    }
  }
  else {
    item.setName("");
    item.setUser("");
  }

  numread += es.read(in, ipthing);
  item.setPassword(es);

  numread += es.read(in, ipthing);
  if(es.length() > 0) {
    data.set(es.get());
    tempdata = data.get();

    item.setNotes(tempdata);
  }
  else {
    item.setNotes("");
  }
  return numread;
}

bool BlowfishLizer::load(Safe &safe, const string &path, const EncryptedString &passphrase, const string &def_user)
{
  //That passkey had better be the same one that came from CheckPassword(...)

#if 0
  try {
#endif
    FILE * in = fopen(path.c_str(), "rb");

    if (in == NULL)
      return false;

    //ClearData(); //Before overwriting old data, but after opening the file...
    unsigned char randstuff[8];
    unsigned char randhash[20];
    unsigned char* salt = new unsigned char[SaltLength];
    unsigned char ipthing[8];
    readHeader(in, randstuff, randhash, salt, ipthing);

    //safe.setRandStuff(randstuff);
    //safe.setRandHash(randhash);

    SafeItem temp;
    //string tempdata;

    // NOTE: the passphrase is decrypted here
    shared_ptr<BlowFish> fish(MakeBlowFish((const unsigned char *)
					   passphrase.get().get(),
					   passphrase.length(),
					   salt, SaltLength));
    EncryptedString es(fish);
    SecuredString data;

    int numread = 0;
    do {
      numread = readEntry(in, temp, es, ipthing, def_user);
      if(numread > 0)
	safe.addItem(temp);
    } while(numread > 0);

    delete [] salt;
    fclose(in);

    return true;
#if 0
  }
  catch(...) {
    // This is probably due to passphrase not being the same as the one
    // passed to checkPassword
    cout << "Exception caught" << endl;
    return false;
  }
#endif
}

bool BlowfishLizer::save(Safe &safe, const string &path, const string &def_user)
{
  FILE *out = fopen(path.c_str(), "wb");
  const EncryptedString &passphrase(safe.getPassPhrase());

  if(out != NULL) {
    unsigned char randstuff[10];
    unsigned char randhash[20];
    for(int i = 0; i < 8; i++)
      randstuff[i] = newrand();
    randstuff[8] = randstuff[9] = 0;

    // NOTE: the passphrase is decrypted here
    SecuredString pphrase(safe.getPassPhrase().get());
    GenRandhash(pphrase.get(), randstuff, randhash);

    fwrite(randstuff, 1, 8, out);
    fwrite(randhash, 1, 20, out);

    // DBoxMain.cpp has a comment about the randomness needing
    // to be checked here, probably a better newrand()
    unsigned char *thesalt = new unsigned char[SaltLength];
    if(thesalt != NULL) {
      int i;

      for(i = 0; i < SaltLength; i++) {
	thesalt[i] = newrand();
      }
      fwrite(thesalt, 1, SaltLength, out);

      unsigned char ipthing[8];
      for(i = 0; i < 8; i++) {
	ipthing[i] = newrand();
      }
      fwrite(ipthing, 1, 8, out);

      // NOTE: the passphrase is decrypted here
      shared_ptr<BlowFish>
	fish(MakeBlowFish((unsigned char *)
			  passphrase.get().get(),
			  passphrase.length(),
			  thesalt, SaltLength));
      EncryptedString es(fish);

      string temp;
      for(Safe::ItemMap::iterator iter = safe.getItems().begin();
	  iter != safe.getItems().end();
	  iter++) {
	if(def_user != iter->second.getUser()) {
	  temp = iter->second.getName();
	  temp += '\xAD';
	  temp += iter->second.getUser();
	}
	else {
	  temp = iter->second.getName();
	  temp += '\xA0';
	}

	//writeCBC(out, temp, passphrase, thesalt, ipthing);
	es.set(temp);
	es.write(out, ipthing);

	//temp = iter->second.getPassword();
	//SecuredString pw(iter->second.getPassword());
	//temp = pw.get();
	//writeCBC(out, temp, passphrase, thesalt, ipthing);
	es.set(iter->second.getPassword().get()); // NOTE: decrypted password
	es.write(out, ipthing);

	temp = iter->second.getNotes();
	//writeCBC(out, temp, passphrase, thesalt, ipthing);
	es.set(iter->second.getNotes());
	es.write(out, ipthing);
      }
      fclose(out);
      delete thesalt;
      return true;
    }

    fclose(out);
    return false;
  }
}
